local GetProfLevel

local function helper(input, ...)
	for i = 1, select("#", ...) do
		local profID = select(i, ...)
		if profID then
			local name, texture, rank, maxRank, numSpells, spelloffset, skillLine = GetProfessionInfo(profID);
			if input == name then
				return rank, maxRank
			end
		end
	end
end

function GetProfLevel(input)
	return helper(input, GetProfessions())
end

local GetHighestSpellIndex = function(index)
	local t = GetNumSpellTabs()
	local n
	while true do
		local name, texture, offset, numSpells = GetSpellTabInfo(t)
		if not name then
			break
		end
		n = offset + numSpells
		t = t + 1
	end
	return n, (index > n)
end

local aName, Trades = ...
local aVer = GetAddOnMetadata(aName, "Version")
local TradesLocale = GetLocale()
local player = UnitName('player')
local _G = _G
local SpellInfo = {}
local TRADESBAR_ADMIN_MAX_BUTTONS = 20

LoadAddOn("EGDebugger")
if EGDebugger then
	EGDebugger:RegisterAddon(aName)
end

TradesBarData = {}

function TradesBar_GetSpellInfo(spellID)
	local spellName = GetSpellInfo(spellID)
	if spellName then
		SpellInfo[spellName] = spellID
	end
	return spellName;
end

TRADES_ALCHEMY = TradesBar_GetSpellInfo(2259)
TRADES_BLACKSMITHING = TradesBar_GetSpellInfo(3100)
TRADES_COOKING = TradesBar_GetSpellInfo(2550)
TRADES_BASIC_CAMPFIRE = TradesBar_GetSpellInfo(818)
TRADES_ENCHANTING = TradesBar_GetSpellInfo(7411)
TRADES_DISENCHANT = TradesBar_GetSpellInfo(13262)
TRADES_ENGINEERING = TradesBar_GetSpellInfo(4036)
TRADES_FIRST_AID = TradesBar_GetSpellInfo(7162)
TRADES_FISHING = TradesBar_GetSpellInfo(7620)
TRADES_HERBALISM = TradesBar_GetSpellInfo(9134)
TRADES_FIND_HERBS = TradesBar_GetSpellInfo(2383)
TRADES_FIND_TREASURE = TradesBar_GetSpellInfo(2481)
TRADES_FIND_FISH = TradesBar_GetSpellInfo(43308)
TRADES_LEATHERWORKING = TradesBar_GetSpellInfo(3104)
TRADES_PICK_LOCK = TradesBar_GetSpellInfo(1804)
TRADES_PICK_POCKET = TradesBar_GetSpellInfo(921)
TRADES_FIND_MINERALS = TradesBar_GetSpellInfo(2580)
TRADES_SMELTING = TradesBar_GetSpellInfo(2656)
TRADES_MINING = TradesBar_GetSpellInfo(2575)
TRADES_SKINNING = TradesBar_GetSpellInfo(8613)
TRADES_TAILORING = TradesBar_GetSpellInfo(3908)
TRADES_JEWELCRAFTING = TradesBar_GetSpellInfo(25229)
TRADES_PROSPECTING = TradesBar_GetSpellInfo(31252)
TRADES_DISARM = TradesBar_GetSpellInfo(1842)
TRADES_FEEDPET = TradesBar_GetSpellInfo(6991)
TRADES_HEARTHSTONE = "item:6948"
TRADES_INSCRIPTION = TradesBar_GetSpellInfo(45358)
TRADES_MILLING = TradesBar_GetSpellInfo(51005)
TRADES_CALLPET1 = TradesBar_GetSpellInfo(0883)
TRADES_CALLPET2 = TradesBar_GetSpellInfo(83242)
TRADES_CALLPET3 = TradesBar_GetSpellInfo(83243)
TRADES_CALLPET4 = TradesBar_GetSpellInfo(83244)
TRADES_CALLPET5 = TradesBar_GetSpellInfo(83245)
TRADES_DISMISSPET = TradesBar_GetSpellInfo(2641)
TRADES_RUNEFORGING = TradesBar_GetSpellInfo(53424)
TRADES_DEATHGATE = TradesBar_GetSpellInfo(50977)
TRADES_ARECALL = TradesBar_GetSpellInfo(556)
TRADES_LOCKPICKING = TradesBar_GetSpellInfo(1809)
TRADES_ARCHAEOLOGY = TradesBar_GetSpellInfo(78670)
TRADES_SURVEY = TradesBar_GetSpellInfo(80451)
TRADES_RUNNING_WILD = TradesBar_GetSpellInfo(87840)
TRADES_TWO_FORMS = TradesBar_GetSpellInfo(68996)
TRADES_INNKEEPER = "item:64488"

BINDING_HEADER_TRADES = aName

BINDING_NAME_TRADES_ALCHEMY = TRADES_ALCHEMY
BINDING_NAME_TRADES_BLACKSMITHING = TRADES_BLACKSMITHING
BINDING_NAME_TRADES_COOKING = TRADES_COOKING
BINDING_NAME_TRADES_ENCHANTING = TRADES_ENCHANTING
BINDING_NAME_TRADES_ENGINEERING = TRADES_ENGINEERING
BINDING_NAME_TRADES_FIRST_AID = TRADES_FIRST_AID
BINDING_NAME_TRADES_LEATHERWORKING = TRADES_LEATHERWORKING
BINDING_NAME_TRADES_SMELTING = TRADES_SMELTING
BINDING_NAME_TRADES_TAILORING = TRADES_TAILORING
BINDING_NAME_TRADES_JEWELCRAFTING = TRADES_JEWELCRAFTING
BINDING_NAME_TRADES_ARCHAEOLOGY = TRADES_ARCHAEOLOGY
BINDING_NAME_TRADES_INSCRIPTION = TRADES_INSCRIPTION

-- Populate _G with locale data
Trades_LoadLocale(TradesLocale)

--[[
local Trades = {
	{
		title = (.*),		-- The name of the skill or item. Use "item:{itemID}" and "GetSpellInfo({spellID})" whenever possible
		panel = [4|5|hide],	-- The panel the admin checkbox will be attached to on the admin frame
		[type = (.*),]		-- The type of button this will be (item, spell, or skill). Default: "spell"
	},
}
--]]

-- "panel = 'hide'" makes it so the button won't show on the config frame.
for _, v in pairs {
	{
		title = TRADES_ALCHEMY,
		panel = 4,
	},
	{
		title = TRADES_BLACKSMITHING,
		panel = 4,
	},
	{
		title = TRADES_COOKING,
		panel = 4,
	},
	{
		title = TRADES_ENCHANTING,
		panel = 4,
	},
	{
		title = TRADES_BASIC_CAMPFIRE,
		panel = 4,
	},
	{
		title = TRADES_DISENCHANT,
		panel = 4,
	},
	{
		title = TRADES_ENGINEERING,
		panel = 4,
	},
	{
		title = TRADES_FIRST_AID,
		panel = 4,
	},
	{
		title = TRADES_FISHING,
		panel = 4,
	},
	{
		title = TRADES_HERBALISM,
		panel = 4,
	},
	{
		title = TRADES_FIND_HERBS,
		panel = 4,
	},
	{
		title = TRADES_FIND_TREASURE,
		panel = 4,
	},
	{
		title = TRADES_FIND_FISH,
		panel = 4,
	},
	{
		title = TRADES_LEATHERWORKING,
		panel = 4,
	},
	{
		title = TRADES_PICK_POCKET,
		panel = 4,
	},
	{
		title = TRADES_PICK_LOCK,
		panel = 4,
	},
	{
		title = TRADES_FIND_MINERALS,
		panel = 4,
	},
	{
		title = TRADES_SMELTING,
		panel = 4,
	},
	{
		title = TRADES_SKINNING,
		panel = 4,
	},
	{
		title = TRADES_TAILORING,
		panel = 4,
	},
	{
		title = TRADES_JEWELCRAFTING,
		panel = 4,
	},
	{
		title = TRADES_MINING,
		panel = 4,
	},
	{
		title = TRADES_PROSPECTING,
		panel = 4,
	},
	{
		title = TRADES_DISARM,
		panel = 4,
	},
	{
		title = TRADES_HEARTHSTONE,
		type = 'item',
		panel = 4,
	},
	{
		title = TRADES_INNKEEPER,
		type = 'item',
		panel = 4,
	},
	{
		title = TRADES_FEEDPET,
		panel = 4,
	},
	{
		title = TRADES_CALLPET1,
		panel = 4,
		callPet = 0883,
	},
	{
		title = TRADES_CALLPET2,
		panel = 4,
		callPet = 83242,
	},
	{
		title = TRADES_CALLPET3,
		panel = 4,
		callPet = 83243,
	},
	{
		title = TRADES_CALLPET4,
		panel = 4,
		callPet = 83244,
	},
	{
		title = TRADES_CALLPET5,
		callPet = 83245,
		panel = 4,
	},
	{
		title = TRADES_DISMISSPET,
		panel = 4,
	},
	{
		title = TRADES_RUNEFORGING;
		panel = 4,
	},
	{
		title = TRADES_DEATHGATE;
		panel = 4,
	},
	{
		title = TRADES_ARECALL;
		panel = 4,
	},
	{
		title = TRADES_INSCRIPTION,
		panel = 4,
	},
	{
		title = TRADES_MILLING,
		panel = 4,
	},
	{
		title = TRADES_ARCHAEOLOGY,
		panel = 4,
	},
	{
		title = TRADES_SURVEY,
		panel = 4,
	},
	{
		title = TRADES_RUNNING_WILD,
		panel = 4,
	},
	{
		title = TRADES_TWO_FORMS,
		panel = 4,
	},
	{
		title = "Random Mount";
		type = "special";
		subtype = "MOUNT";
		panel = 4,
	},
	{
		title = "Random Pet";
		type = "special";
		subtype = "CRITTER";
		panel = 4,
	},
} do tinsert(Trades, v) end

local buttonsEnabled = true
local function GetInfo(self)
	if not self then
		return false
	end
	if self.info then
		return self.info
	end
	for _, v in pairs(Trades) do
		if v.title == self.name then
			self.info = v
			return v
		end
	end
end 

local AdminPanels = {};
local original_heights = {}

local TRADESBAR_ENABLED = true;
local TradesBar_UpdateInterval = 1.0;
local TradesBar_TooltipRanksText = {}

local color = {
	RED			= "ff0000"; -- current rank <= 25% of max
	ORANGE		= "ff7f00"; -- current rank <= 50% and > 25% of max
	YELLOW		= "ffff00"; -- current rank <= 75% and > 50% of max
	GREEN		= "00ff00"; -- current rank <= 100% and > 75% of max
}

local BagAddons = {
	BaudBag = function()
		BaudBagAutoOpenSet(1);
		BaudBagAutoOpenSet(2);
	end,
	OneBag = function()
		OneBagFrame:Show()
	end,
	ArkInventory = function()
		OpenAllBags(true)
	end,
	EngBags = function()
		EngBank_OpenOrClose("open")
	end,
}

local function ToggleButtons(mode)
	if mode == "Disable" then
		buttonsEnabled = true
	elseif mode == "Enable" then
		buttonsEnabled = false
	end
	local method = buttonsEnabled and "Disable" or "Enable"
	local alpha = buttonsEnabled and .1 or 1
	local button
	for i, v in pairs(Trades) do
		button = v.button
		if button then
			button[method](button)
			button:EnableMouse(not buttonsEnabled)
			button:SetAlpha(alpha)
		end
		TradesBar:EnableMouse(buttonsEnabled)
	end
	buttonsEnabled = not buttonsEnabled
end

function TradesBar_AddSpellOrItem()
	if not CursorHasItem() and not CursorHasSpell() then
		return
	end
	local infoType, info1, info2, info3 = GetCursorInfo()
	ClearCursor()
	if infoType == 'item' then
		local itemName = GetItemInfo(info1)
		if itemName then
			TradesBar_AddItem(itemName)
		end
	elseif infoType == BOOKTYPE_SPELL then
		local _, ob = GetHighestSpellIndex(info1)
		if info1 == 0 or ob then -- As of 3.0.2 GetCursorInfo() returns "spell", 0, "spell" for mini-pets and mounts. Alert the user and return.
			return DEFAULT_CHAT_FRAME:AddMessage("Invalid spell ID, this is probably a mount or pet. These are added automatically.", 1.0, 0.0, 0.25);
		end
		local spellName = GetSpellBookItemName(info1, BOOKTYPE_SPELL);
		TradesBar_AddSpell(spellName)
	end
	TradesBar_Load()
	TradesBarConfig_OnLoad(TradesBarConfig)
end

function TradesBar_AddSpell(spell)
	local spellName = GetSpellInfo(spell)
	if not spellName then
		local highSpellIndex, ob = GetHighestSpellIndex(0)
		for i = 1, highSpellIndex do
			spellName = GetSpellBookItemName(i, BOOKTYPE_SPELL);
			if spellName then
				break
			end
		end
	end
	if (EGDebugger and EGDebugger:AddonRegistered(aName)) then
		EGDebugger:PrintToStdOut(aName, "Found |cffffffff'"..spellName.."'|r at slot |cffffffff'"..i.."'", 0.1, 1.0, 0.0);
	end
	local spellTable = {
		title = spellName,
		panel = 5,
	}
	if not TradesBar_InTable(Trades, spellName) then
		table.insert(Trades, spellTable); -- Add to "Trades" for real-time adding of skills.
		table.insert(TradesBarCustomSpells, spellTable);
		TradesBarData["skills"][spellName] = TradesBarData["skills"][spellName] or {}
		TradesBarData["skills"][spellName].visible = "true"
		AdminPanels[5] = AdminPanels[5] or {};
		table.insert(AdminPanels[5], spellTable);
	end
end

function TradesBar_AddItem(item)
	-- The item *must* be in your local cache for this to work.
	local itemName, itemLink = GetItemInfo(item)
	if itemLink then
		local itemWord, itemId
		local itemString = strmatch(itemLink, "^|c%x+|H(.+)|h%[.+%]")
		if itemString then
			itemId = strmatch(itemString, "item:(%d+):.+")
-- 			DEFAULT_CHAT_FRAME:AddMessage('item:'..itemId)
			local itemTable = {
				title = 'item:'..itemId,
				type = 'item',
				panel = 5,
			}
			if TradesBar_InTable(Trades, 'item:'..itemId) == false then
				table.insert(Trades, itemTable); -- Add to "Trades" for real-time adding of items.
				table.insert(TradesBarCustomSpells, itemTable);
				TradesBarData["skills"]['item:'..itemId] = TradesBarData["skills"]['item:'..itemId] or {}
				TradesBarData["skills"]['item:'..itemId].visible = "true"
				AdminPanels[5] = AdminPanels[5] or {};
				table.insert(AdminPanels[5], itemTable);
			end
		end
	end
end

function TradesBar_GetSpecialButton(ptype, buttonName)
	local spellID, icon
	local button = CreateFrame("Button", buttonName, TradesBar, "TradesBarButtonTemplate");
	button.newMountID = 1
	button.companions = GetNumCompanions(ptype)
	button:RegisterEvent('COMPANION_LEARNED');
	button:RegisterEvent('COMPANION_UNLEARNED');
	
	button:SetScript("OnLeave", function(self)
		self.myTooltip = nil
		GameTooltip:Hide()
	end)
	
	button:SetScript("OnEvent", function(self, ...)
		self.companions = GetNumCompanions(ptype)
	end)
	
	button:SetScript("OnEnter", function(self)
		self.myTooltip = 1
		GameTooltip_SetDefaultAnchor(GameTooltip, self)
		local colors = ChatTypeInfo["SYSTEM"]
		if not self.spinning then
			GameTooltip:SetHyperlink("spell:" .. spellID)
			GameTooltip:AddLine("Left-Click to start spinning without summoning or summon the current companion at the time of the click.", colors.r, colors.g, colors.b, colors.id)
			GameTooltip:AddLine(" ")
			GameTooltip:AddLine("Right-Click to stop spinning without summoning or bypass the spinning and summon the companion.", colors.r, colors.g, colors.b, colors.id)
			GameTooltip:AddLine(" ")
			GameTooltip:AddLine("== Your " .. (ptype == "MOUNT" and "Mounts" or "Pets") .. " ==")
			for i = 1, self.companions do
				local _, name = GetCompanionInfo(ptype, i)
				if name then
					GameTooltip:AddLine(name)
				end
			end
		else
			GameTooltip:AddLine("Spinning...", colors.r, colors.g, colors.b, colors.id)
		end
		GameTooltip:Show()
	end)
	
	button:SetScript("OnUpdate", function(self, elapsed)
		if self.spinning then
			self.newMountID = (self.newMountID % self.companions) + 1
			self.timeSince = 0
			if GameTooltip:IsShown() and self.myTooltip then
				self:GetScript("OnEnter")(self)
			end
		end
		if self.prevMountID ~= self.newMountID then
			self:ChangeMountID(self.newMountID)
			self.prevMountID = self.newMountID
		end
	end)
	button:SetScript("OnClick", function(self, button)
		if not IsShiftKeyDown() and not IsControlKeyDown() and not IsAltKeyDown() then
			-- Covers Left-Click while spinning and Right-Click while not spinning
			if (self.spinning and button == "LeftButton") or (not self.spinning and button == "RightButton") then
				self:ChangeMountID(self.newMountID)
				CallCompanion(ptype, self.newMountID)
				self.spinning = nil
				if GameTooltip:IsShown() and self.myTooltip then
					self:GetScript("OnEnter")(self)
				end
			-- Covers Right-Click while spinning.
			elseif button == "RightButton" then
				self.spinning = nil
				if GameTooltip:IsShown() and self.myTooltip then
					self:GetScript("OnEnter")(self)
				end
			-- Covers Left-Click while not spinning
			else
				self.spinning = 1
			end
		end
	end)
	
	function button:ChangeMountID(newMountID)
		spellID, icon = select(3, GetCompanionInfo(ptype, newMountID))
		if not InCombatLockdown() then
			self:SetID(newMountID)
			_G[self:GetName() .. "Icon"]:SetTexture(icon)
			self.texture = icon
		end
	end
	
	button:ChangeMountID(button.newMountID)
	return button
end

StaticPopupDialogs["TradesBar_LoadDefault"] = {
	text = format(TRADES_TEXT_LOAD_DEFAULT, '"'..YES..'"'),
	button1 = YES,
	button2 = CANCEL,
	OnAccept = function(this)
		TradesBarData = nil;
		TradesBarCustomSpells = {};
		ReloadUI();
	end,
	OnCancel = function(this)
		this:Hide();
	end,
	hideOnEscape = 1,
	timeout = 0,
	exclusive = 1,
	whileDead = 1
}

function TradesBar_OnLoad(this)
	-- Hooked for debugging
--	hooksecurefunc('SecureActionButton_OnClick',
--		function (self, button)
--			local spell = SecureButton_GetModifiedAttribute(self, "spell", button);
--			DEFAULT_CHAT_FRAME:AddMessage('Click accepted for button '..spell..'('..button..')');
--		end
--	);
	TradesBar_TimeSinceLastUpdate = 0;
	this:RegisterEvent('ADDON_LOADED');
	this:RegisterEvent("PLAYER_LOGIN");
    this:RegisterEvent("PET_BATTLE_OPENING_START");
    this:RegisterEvent("PET_BATTLE_OVER");
	this:EnableMouse(false)
	
	SLASH_TRADESBAR1 = "/tradesbar";
	SLASH_TRADESBAR2 = "/tbar";
	SLASH_TRADESBAR3 = "/tb";
	SlashCmdList["TRADESBAR"] = function(msg)
		TradesBar_SlashHandler(string.lower(msg));
	end
	tinsert(UISpecialFrames, "TradesBarConfig"); 
end

function TradesBar_HelpMsg()
	for i = 1, 3 do
		DEFAULT_CHAT_FRAME:AddMessage(_G["TRADES_HELP_MESSAGE"..i]);
	end
end

function TradesBar_SlashHandler(msg, mode)
	if (string.find(msg, "help")) then
		TradesBar_HelpMsg();
	elseif (string.find(msg, "toggle")) then
		TradesBar_Toggle();
	elseif (string.find(msg, "default")) then
		StaticPopup_Show('TradesBar_LoadDefault');
	elseif EGDebugger and (string.find(msg, "debug toggle")) then
		EGDebuggerMessages["enabled"][aName] = not EGDebuggerMessages["enabled"][aName]
	elseif EGDebugger and (string.find(msg, "debug off")) then
		EGDebuggerMessages["enabled"][aName] = nil
	elseif EGDebugger and (string.find(msg, "debug on")) then
		EGDebugger:RegisterAddon(aName)
	elseif (string.find(msg, "debug")) then
		TradesBarDump()
	elseif (string.find(msg, "admin") or string.find(msg, "config")) then
		if(	TradesBarConfig:IsVisible() ) then
			TradesBarConfig:Hide();
		else
			TradesBarConfig:Show();
		end
	elseif msg == "move" then
		ToggleButtons(mode)
	else
		TradesBar_HelpMsg();
	end
end

function StartTradesBar(self)
	if EGDebugger and EGDebugger:AddonRegistered(aName) then
		TradesBarConfig:Show();
	end
	TradesBar_UpdateCooldowns();
	if CustomTradesOrder then
		Trades = CustomTradesOrder;
	end
	for i, v in pairs(Trades) do
		if v.title then
			TradesBarData["skills"] = TradesBarData["skills"] or {}
			if not TradesBarData["skills"][v.title] then
				TradesBarData["skills"][v.title] = {}
			end
			if (not TradesBarData["skills"][v.title].visible) then
				TradesBarData["skills"][v.title].visible = "true";
			end
		end
	end
	TradesBar_ScanPets(1)
	TradesBar_Load(self)
	TradesBar_Scale();
	TradesBar_Update();
end

function TradesBar_InTable(haystack, needle)
	if not haystack then
		return false
	end
	for i, v in pairs(haystack) do
		if v.title == needle then
			return i
		end
	end
	return false
end

do
	local mountTypes = {"MOUNT", "CRITTER"}
	function TradesBar_ScanPets(onLoad)
		-- Mounts and Pets
		local creatureID, creatureName, spellID, icon, active
		for _, ctype in pairs(mountTypes) do
			if TradesBarData["settings"]["hide_all_" .. ctype] == "false" or onLoad then
				for index = 1, GetNumCompanions(ctype) do
					creatureID, creatureName, spellID, icon, active = GetCompanionInfo(ctype, index)
					if creatureName and TradesBar_InTable(Trades, creatureName) == false then
						table.insert(Trades, {
							title = creatureName,
							type = "macro",
							macrotext = "/run CallCompanion('" .. ctype .. "'," .. index .. ")",
							texture = icon,
							panel = 4,
							mountOrPet = ctype,
							spellID = spellID,
						})
						if not TradesBarData["skills"][creatureName] then
							TradesBarData["skills"][creatureName] = {}
						end
						if not TradesBarData["skills"][creatureName].visible then
							TradesBarData["skills"][creatureName].visible = "true"
							TradesBarData["skills"][creatureName].buttonType = ctype
						end
						if TradesBarData["settings"]["hide_all_"..ctype] == "true" then
							TradesBarData["skills"][creatureName].visible = "false"
						end
					end
				end
			end
		end
	end
end

function TradesBar_MoveFrame(frame, point1, rel, point2, x, y)
	if frame and x ~= nil and y ~= nil and point1 ~= nil and point2 ~= nil and rel ~= nil then
		frame:ClearAllPoints()
		frame:SetPoint(point1, rel, point2, x, y)
	end
end

function TradesBar_Variables()
	if not TradesBarCustomSpells then
		TradesBarCustomSpells = {}
	end
	local pname = player
	if (pname ~= UNKNOWNBEING and pname ~= UNKNOWNOBJECT and pname and not playerName) then
		playerName = pname.." of "..GetCVar("realmName");
	end
	if not TradesBarData then
		TradesBarData = {}
	end
	if not TradesBarData["skills"] then -- seems a little unnecessary
		TradesBarData["skills"] = {}
	end
	if not TradesBarData["settings"] then
		TradesBarData["settings"] = {
			["hidebar"] = "false",
			["horizontal"] = "true",
			["order"] = "az",
			["scale"] = 1,
		}
	end
	if TradesBarData[playerName] then
		TradesBarData["settings"] = {
			["hidebar"] = TradesBarData[playerName]["hidebar"],
			["horizontal"] = TradesBarData[playerName]["horizontal"],
			["order"] = TradesBarData[playerName]["order"],
			["scale"] = TradesBarData[playerName]["scale"],
		}
		TradesBarData["skills"] = {}
		for k in pairs(TradesBarData[playerName]) do
			if k ~= "version" and k ~= "hide_all_MOUNT" and k ~= "hide_all_CRITTER" and k ~= "hidebar" and k ~= "horizontal"  and k ~= "order"  and k ~= "scale" and k ~= "updated" and k ~= "left" and k ~= "top" then
				TradesBarData["skills"][k] = TradesBarData[playerName][k]
			end
			TradesBarData[playerName][k] = nil
		end
		TradesBarData[playerName] = nil
	end
	for k, v in pairs(TradesBarData["skills"]) do
		if TradesBarData["skills"][k] == "true" or TradesBarData["skills"][k] == "false" then
			TradesBarData["skills"][k] = {
				["visible"] = v,
				["key"] = nil,
			}
		end
	end
	TradesBar_MoveFrame(TradesBarConfigPanel1AddSpellButtonText1, 'TOPLEFT', TradesBarConfigPanel1AddSpellButton, 'TOPLEFT', -1 * TradesBarConfigPanel1AddSpellButtonText1:GetWidth() - 10, 2)
	TradesBar_MoveFrame(TradesBarConfigPanel1AddSpellButtonText2, 'TOPLEFT', TradesBarConfigPanel1AddSpellButton, 'TOPLEFT', -1 * TradesBarConfigPanel1AddSpellButtonText2:GetWidth() - 10, -10)
	TradesBarBindingFrame.playername = playerName
end

local eventsFired = {}
local in_loading_screen
local delayTimer = .1
local delay, delayFrame = delayTimer
function TradesBar_OnEvent(self, event, ...)
	-- don't process events while inside loading screens 
	-- 1. makes for much faster loading 
	-- 2. we avoid errors with GetSpellBookItemTextuer and GetSpellCooldown on spells while spellbook is not yet loaded.
	if in_loading_screen then
		return
	end
	eventsFired[event] = true
	if (event == "PET_BATTLE_OPENING_START") then
		self.PET_BATTLE_OPENING_START = true
		self.wasVisible = self:IsVisible();
		if self.wasVisible then
			self:Hide();
		end
	elseif (event == "PET_BATTLE_OVER") then
		self.PET_BATTLE_OPENING_START = false
		if self.wasVisible then
			self:Show();
		end
	elseif (event == 'ADDON_LOADED') then
		if (...) == "TradesBar" then
			TradesBar_Variables()
		end
		if (...) == 'FishingBuddy' then
			TradesBar_is_fishingbuddy = true;
		end
	elseif (event == "PLAYER_LOGIN") then
		if CustomTradesBar and type(CustomTradesBar) == 'function' then
			CustomTradesBar();
		end
		-- Add the custom buttons
		if TradesBarCustomSpells then
			for i in ipairs(TradesBarCustomSpells) do
				if TradesBar_InTable(Trades, TradesBarCustomSpells[i].title) == false then
					if TradesBarCustomSpells[i].name then
						TradesBarCustomSpells[i].title = TradesBarCustomSpells[i].name
						TradesBarCustomSpells[i].name = nil
					end
					if (TradesBarCustomSpells[i].type == 'item' and GetItemInfo(TradesBarCustomSpells[i].title)) or TradesBarCustomSpells[i].type ~= 'item' then
						table.insert(Trades,
							{
								title = TradesBarCustomSpells[i].title,
								panel = 5,
								type = TradesBarCustomSpells[i].type,
							}
						);
					end
				else
					TradesBarCustomSpells[i] = nil;
				end
			end
		end
		-- Add the custom buttons
		if CustomTradesBarFromFile then
			for i in ipairs(CustomTradesBarFromFile) do
				if TradesBar_InTable(Trades, CustomTradesBarFromFile[i].title) == false then
					if (CustomTradesBarFromFile[i].type == 'item' and GetItemInfo(CustomTradesBarFromFile[i].title)) or CustomTradesBarFromFile[i].type ~= 'item' then
						table.insert(Trades,
							{
								title = CustomTradesBarFromFile[i].title,
								panel = 5,
								type = CustomTradesBarFromFile[i].type,
								texture = CustomTradesBarFromFile[i].texture,
								macrotext = CustomTradesBarFromFile[i].macrotext,
								tooltiptext = CustomTradesBarFromFile[i].tooltiptext,
							}
						);
					end
				else
					CustomTradesBarFromFile[i] = nil;
				end
			end
		end
		StartTradesBar(self);
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_LEAVING_WORLD")
		self:RegisterEvent("SPELL_UPDATE_COOLDOWN");
		self:RegisterEvent('COMPANION_UPDATE');
		self:RegisterEvent("SPELLS_CHANGED");
		self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
		self:RegisterEvent("UNIT_AURA");
	elseif (event == "PLAYER_LEAVING_WORLD") then
		in_loading_screen = true
	elseif (event == "PLAYER_ENTERING_WORLD") then
		in_loading_screen = false
	elseif (event == "COMPANION_UPDATE") then
		TradesBar_ScanPets(1)
		TradesBar_Load(self)
		TradesBarConfig_OnShow(TradesBarConfig)
	elseif (event == "SPELLS_CHANGED" or (event == "UNIT_AURA" and ... == "player") or event == "ACTIVE_TALENT_GROUP_CHANGED") then
		TradesBar_Load(self)
		delay = delayTimer
		if not delayFrame then
			delayFrame = CreateFrame("Frame")
			delayFrame:SetScript("OnUpdate", function(self, e)
				delay = delay - e
				if delay <= 0 then
					TradesBar_Hide(self)
					TradesBar_Update(self, true);
				end
			end)
		end
		delayFrame:Show()
	elseif (event == "SPELL_UPDATE_COOLDOWN") then
		TradesBar_UpdateCooldowns();
	end
end

function TradesBar_ImplodeList(char, ...)
	local ret = select(1, ...)
	for i = 2, select("#", ...) do
		ret = ret .. char .. " " .. select(i, ...)
	end
	return ret
end

local queueframe = CreateFrame("Frame")
function queueframe:Load()
	self.queueitems = {
		-- this is the proper order
		TradesBar_Load = false, -- executes secure code -- cannot be done in combat
		TradesBar_Update = false, -- executes secure code -- cannot be done in combat
		TradesBar_O = false, -- executes secure code -- cannot be done in combat
		TradesBar_Toggle = false, -- shows/hides a secure frame -- cannot be done in combat
		TradesBar_Hide = false, -- shows/hides a secure frame -- cannot be done in combat
		TradesBar_Scale = false, -- SetScale on secure frames -- cannot be done in combat
	}
	self:SetScript("OnUpdate", function(self, elapsed)
		if not InCombatLockdown() then
			for func, bool in pairs(self.queueitems) do
				if bool == true then
					_G[func]()
					self.queueitems[func] = false
				end
			end
			self:Hide()
		end
	end)
end
queueframe:Load()

do
	local OpenBagItems = {
		[TRADES_DISENCHANT] = true;
		[TRADES_FEEDPET] = true;
		[TRADES_PICK_LOCK] = true;
		[TRADES_PROSPECTING] = true;
		[TRADES_MILLING] = true;
	}
		
	function TradesBar_Load(self)
		if not eventsFired["PLAYER_LOGIN"] then
			return
		end
		if InCombatLockdown() then
			queueframe.queueitems.TradesBar_Load = true
			queueframe:Show()
			return
		end
		for id, v in pairs(Trades) do
			if not v.button then
				if v.type == "special" then
					v.button = TradesBar_GetSpecialButton(v.subtype, 'TradesBarButton'..id)
				else
					v.button = CreateFrame('Button', 'TradesBarButton'..id, TradesBar, 'TradesBarButtonTemplate');
				end
			end
			v.button.name = v.title
			v.type = v.type or BOOKTYPE_SPELL
			v.spellID = v.spellID or SpellInfo[v.title];
-- 			DEFAULT_CHAT_FRAME:AddMessage(v.title..': '..tostring(v.type))
			if v.type == 'item' then
				local _, itemLink = GetItemInfo(v.title)
-- 				DEFAULT_CHAT_FRAME:AddMessage(tostring(itemLink))
				v.itemLink = itemLink;
			end
			v.button:RegisterForClicks('AnyUp');
			-- If it has a panel, then we need to tell the script to add it and where to add it to the config frame.
			if EGDebugger and EGDebugger:AddonRegistered(aName) then
				EGDebugger:PrintToStdOut(aName, 'Adding AdminPanel['..tostring(v.panel)..']['..id..']')
			end
			if not v.adminCheck then
				v.adminCheck = 1
				if not TradesBar_InTable(AdminPanels[4], v.title) and not TradesBar_InTable(AdminPanels[5], v.title) then
					if v.panel then
						AdminPanels[v.panel] = AdminPanels[v.panel] or {};
						table.insert(AdminPanels[v.panel], {
							title = v.title;
							panel = v.panel;
							type = v.type;
							subtype = v.subtype;
						})
					else
						AdminPanels[5] = AdminPanels[5] or {};
						table.insert(AdminPanels[5], {
							title = v.title;
							panel = v.panel;
							type = v.type;
						})
					end
				end
			end
			if v.type ~= 'skill' and v.type ~= 'macro' and v.type ~= "special" then
				-- Required for SecureActionButtonTemplate
				v.button:SetAttribute("type", v.type);
				v.button:SetAttribute("shift-" .. v.type .. "2", ATTRIBUTE_NOOP);
				v.button:SetAttribute("ctrl-" .. v.type .. "2", ATTRIBUTE_NOOP);
				v.button:SetAttribute("alt-" .. v.type .. "2", ATTRIBUTE_NOOP);
				v.button:SetAttribute(v.type, v.title);
			elseif v.type == 'macro' then
				v.button:SetAttribute("type", v.type);
				v.button:SetAttribute("shift-macro2", ATTRIBUTE_NOOP);
				v.button:SetAttribute("ctrl-macro2", ATTRIBUTE_NOOP);
				v.button:SetAttribute("alt-macro2", ATTRIBUTE_NOOP);
			elseif v.type == "special" then -- beats wrapping it in another if block
				-- do nothing
			else
				v.button:SetScript("OnClick",
					function(self, ...)
						local info = GetInfo(self)
						if not info then
							return
						end
						CastSpellByName(info.title) -- does not work for "actions"
					end
				)
			end
			if OpenBagItems[v.title] == true then
				v.button:SetScript('OnMouseUp',
					function(self, ...)
						for addon, func in pairs(BagAddons) do
							if IsAddOnLoaded(addon) then
								return func()
							end
						end
						-- If the above didn't return, use the default method.
						OpenAllBags(self);
					end
				)
			end
		end
		TradesBar_Update()
	end
end

function TradesBar_Scale()
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_Scale = true
		queueframe:Show()
		return
	end
	TradesBarData["settings"] = TradesBarData["settings"] or {}
	TradesBar:SetScale(TradesBarData["settings"]["scale"] or 1);
end

-- Convienience functions
function TradesBar_HasSkill(skill)
	if GetSpellInfo(skill) then
		return true;
	else
		return false;
	end
end

function TradesBar_HasItem(button)
	local info = GetInfo(button)
	if not info then
		return
	end
	if info.type ~= 'item' then
		return
	end
	local searchItem = GetItemInfo(info.title)
	if GetItemCount(searchItem) > 0 then
		return true
	end
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local itemlink = GetContainerItemLink(bag, slot)
			if itemlink then
				local item = GetItemInfo(itemlink)
				if searchItem == item then
					return true
				end
			end
		end
	end
end

function TradesBar_O()
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_O = true
		queueframe:Show()
		return
	end
	TradesBar:SetWidth(0)
	TradesBar:SetHeight(0)
	-- Default to horizontal
	local prev, button, point, relative, x, y, height, width
	x, y = 0, 0
	height, width = 28, 28
	for i, v in pairs(Trades) do
		if v.title then
			v.button:ClearAllPoints();
			-- Default to visible
			TradesBarData["skills"][v.title] = TradesBarData["skills"][v.title] or {}
			if not TradesBarData["skills"][v.title].visible then
				TradesBarData["skills"][v.title].visible = "true"
			end
			TradesBar_visableCheck(v.button);
			if not v.hide_button and not (v.type == 'item' and GetItemCount(v.title) == 0) and ((v.type and ((v.type == 'item' and TradesBar_HasItem(v.button)) or v.type == 'macro' or v.type == 'special')) or TradesBar_HasSkill(v.title)) then
				prev = button
				button = v.button;
				if v.type == 'macro' then
					button:SetAttribute('macro1', false)
					button:SetAttribute('macrotext', button.macrotext);
				end
				if not button.companions or button.companions > 0 then
	-- 				DEFAULT_CHAT_FRAME:AddMessage('Button '..tostring(v.title));
	-- 				DEFAULT_CHAT_FRAME:AddMessage('PrevButton '..prev);
					if (TradesBarData["settings"]["horizontal"] == "true" and TradesBarData["settings"]["order"] == "az") then
						point, relative = "TOPLEFT", "TOPLEFT";
						if prev then
							x = x + 28
							y = y + 0
							width = width + 28
						end
					elseif (TradesBarData["settings"]["horizontal"] == "true" and TradesBarData["settings"]["order"] == "za") then
						point, relative = "TOPRIGHT", "TOPRIGHT";
						if prev then
							x = x - 28
							y = y + 0
							width = width + 28
						end
					elseif (TradesBarData["settings"]["horizontal"] == "false" and TradesBarData["settings"]["order"] == "az") then
						point, relative = "TOPRIGHT", "TOPRIGHT";
						if prev then
							x = x + 0
							y = y - 28
							height = height + 28
						end
					elseif (TradesBarData["settings"]["horizontal"] == "false" and TradesBarData["settings"]["order"] == "za") then
						point, relative = "BOTTOMLEFT", "BOTTOMLEFT";
						if prev then
							x = x + 0
							y = y + 28
							height = height + 28
						end
					else
						point, relative = "TOPLEFT", "TOPLEFT";
						if prev then
							x = x + 28
							y = y + 0
							width = width + 28
						end
					end
					button:SetPoint(point, 'TradesBar', relative, x, y);
					TradesBar:SetWidth(width)
					TradesBar:SetHeight(height)
				end
			end
		end
	end
end

function TradesBar_Hide(f)
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_Hide = true
		queueframe:Show()
		return
	end
	local frame = f or TradesBar
	if frame then
		frame.wasVisible = frame:IsVisible();
		if frame.wasVisible then
			frame:Hide()
		end
	end
end

function TradesBar_Toggle(f)
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_Toggle = true
		queueframe:Show()
		return
	end
	local frame = f or TradesBar;
	if frame then
		frame.wasVisible = frame:IsVisible();
		if frame.wasVisible then
			frame:Hide();
		elseif not TradesBar.PET_BATTLE_OPENING_START then
			frame:Show();
		end
	end
end

function TradesBar_FishingBuddy_Over()
	TradesBar_Toggle(FishingBuddyFrame)
end

function TradesBar_visableCheck(button)
	local info = GetInfo(button)
	if not info then
		return false
	end
	info.hide_button = true;
	local value = info.title
	if not value then
		return
	end
	-- Special conditions
	if (value == TRADES_GNOME_ENGINEERING or value == TRADES_GOBLIN_ENGINEERING) then
		value = TRADES_ENGINEERING;
	end
	if (value == TRADES_DRAGONSCALE_LEATHERWORKING or value == TRADES_ELEMENTAL_LEATHERWORKING or value == TRADES_TRIBAL_LEATHERWORKING) then
		value = TRADES_LEATHERWORKING;
	end
	if (value == TRADES_SHADOWEAVE_TAILORING or value == TRADES_SPELLFIRE_TAILORING) then
		value = TRADES_TAILORING;
	end
	-- I really hate doing it this way, but it seems necessary. Grr...
	TradesBarData = TradesBarData or {}
	TradesBarData["skills"] = TradesBarData["skills"] or {}
	TradesBarData["skills"][value] = TradesBarData["skills"][value] or {}
	if (TradesBarData["skills"][value].visible == "true" or TradesBarData["skills"][value].visible == nil) then
		info.hide_button = false;
	end
end

function TradesBar_Update(self, refresh)
	if InCombatLockdown() then
		queueframe.queueitems.TradesBar_Update = true
		queueframe:Show()
		return
	end
	if (TradesBarData["settings"] and TradesBarData["settings"]["hidebar"] == "false") then
		local i = 0;
		for key, v in pairs(Trades) do
			if v.button then
				v.button:Hide();
			end
		end
		local height = 0
		for key, v in pairs(Trades) do
			local id = not refresh and v.spellNum
			v.isFlyout = nil
			if not id then
				id = TradesBar_GetSpellID(v.title);
			end
			v.spellNum = id
			TradesBar_visableCheck(v.button);
			if (not v.hide_button) then
				if (id or v.callPet) then
					if not id then
						id = v.callPet
						v.isFlyout = 1
					end
-- 					DEFAULT_CHAT_FRAME:AddMessage('Found '..v.title);
					i = i + 1;
					local button = v.button;
					if button then
						local _, texture, retOK
						if v.isFlyout then
							_, _, texture = GetSpellInfo(v.title);
						else
							local _, ob = GetHighestSpellIndex(id)
							if not ob then
								texture = GetSpellBookItemTexture(id, BOOKTYPE_SPELL)
							end
						end
						if not texture then
							_, _, texture = GetSpellInfo(v.title)
						end
						if --[[still]] not texture then
							_, _, _, _, _, _, _, _, _, itemTexture = GetItemInfo(v.title)
						end
						_G[button:GetName().."Icon"]:SetTexture(texture);
						v.id = id;
						button.id = id;
						button.texture = texture;
						button:Show();
						height = height + 28
					end
				elseif v.type == 'item' then
-- 					DEFAULT_CHAT_FRAME:AddMessage('Found '..v.title);
					i = i + 1;
					local button = v.button;
					local _, _, _, _, _, _, _, _, _, texture = GetItemInfo(v.title);
					_G[button:GetName().."Icon"]:SetTexture(texture);
					button.texture = texture;
					button:Show();
					height = height + 28
				elseif v.type == 'macro' then
					i = i + 1;
					local button = v.button;
					_G[button:GetName().."Icon"]:SetTexture(v.texture);
					button.texture = v.texture;
					button.macrotext = v.macrotext;
					button.tooltiptext = v.tooltiptext;
					button:Show();
					height = height + 28
				elseif v.type == 'special' then
					local button = v.button;
					if button.companions > 0 then
						i = i + 1;
						button:Show();
						height = height + 28
					end
				end
			end
		end
		if TradesBarData["settings"]["horizontal"] == 'true' then
			TradesBar:SetWidth(height)
			TradesBar:SetHeight(28)
		else
			TradesBar:SetWidth(28)
			TradesBar:SetHeight(height)
		end
		if (i > 0) and not TradesBar.PET_BATTLE_OPENING_START then
			TradesBar:Show();
		end
		TradesBar_O();
	end
end

function TradesBar_UpdateCooldowns()
	for i, v in pairs(Trades) do
		if v.button then
			if v.id then
				local cooldown = _G[v.button:GetName().."Cooldown"];
				local start, duration, enable, retOK
				retOK, start, duration, enable = pcall(GetSpellCooldown, v.id, v.spellID ~= v.id and BOOKTYPE_SPELL or nil); -- pcall this so we can get the actual params instead of a "Usage: " error.
				if retOK and (start and duration and start > 0 and duration > 0) then
					CooldownFrame_SetTimer(cooldown, start, duration, enable);
				elseif not retOK and (EGDebugger and EGDebugger:AddonRegistered(aName)) then
					EGDebugger:PrintToStdOut(aName, " error: GetSpellCooldown("..tostring(v.id)..","..tostring(v.spellID ~= v.id and BOOKTYPE_SPELL or nil)..")".." v.spellID:"..tostring(v.spellID).."\n"..start)
				end
			elseif v.type == 'item' and v.itemLink then
				local cooldown = _G[v.button:GetName().."Cooldown"];
				local itemid = v.itemLink:match("item:(.*):")
				if itemID then
					local start, duration, enable = GetItemCooldown(itemID);
					if (start > 0 and duration > 0) then
						CooldownFrame_SetTimer(cooldown, start, duration, enable);
					end
				end
			end
		end
	end
end

function TradesBar_GetSpellID(spell, rank)
	local n = GetHighestSpellIndex(0)
	local spellID;
	local highestRank;
	for i = 1, n do
		local spellName, spellRank = GetSpellBookItemName(i, BOOKTYPE_SPELL);
		if (not spellName) then
			break;
		end
		if (spellName == spell) then
			if (rank) then
				if (spellRank == rank) then
					if (EGDebugger and EGDebugger:AddonRegistered(aName)) then
						EGDebugger:PrintToStdOut(aName, "Found |cffffffff'"..spell.." ("..rank..")'|r at slot |cffffffff'"..i.."'", 0.1, 1.0, 0.0);
					end
					return i;
				end
			else
				spellID = i;
				highestRank = spellRank;
			end
		end
	end			
	if (EGDebugger and EGDebugger:AddonRegistered(aName)) then
		EGDebugger:PrintToStdOut(aName, "Could not find |cffffffff'"..spell.." ("..rank..")'", 1.0, 0.1, 0.0);
	end
	return spellID, highestRank;
end

function TradesBarButtonTemplate_OnEnter(self)
	local skillColor, skillDiffPercent, _ = '', 0;
	local x = self:GetRight()
	local info = GetInfo(self)
	if not info then
		return
	end
	if ( x >= ( GetScreenWidth() / 2 ) ) then
		GameTooltip:SetOwner(self, "ANCHOR_LEFT");
	else
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
	end
	if info.type ~= "item" and not info.spellID then
		info.spellID = SpellInfo[info.title] or TradesBar_GetSpellIDByName(info.title)
	end
	if info.mountOrPet then
		GameTooltip:SetHyperlink("spell:"..info.spellID);
		GameTooltip:Show();
	elseif (info.id) then
		local skillName
		if info.callPet or info.id == info.spellID then
			skillName = GetSpellInfo(info.id);
			self.UpdateTooltip = SpellFlyoutButton_SetTooltip
			self.spellID = info.spellID
			SpellFlyoutButton_SetTooltip(self)
			return
		else
			local _, ob = GetHighestSpellIndex(info.id)
			if not ob then
				skillName = GetSpellBookItemName(info.id, BOOKTYPE_SPELL)
			end
		end
		tmpSkillName = TradesBar_ProperSkillName(skillName);
		if tmpSkillName ~= skillName then
			self.spellID = info.spellNum
			if not self.spellID then
				self.spellID = TradesBar_GetSpellID(tmpSkillName);
			end
			skillName = tmpSkillName;
		else
			self.spellID = info.id;
		end
		if info.spellID and GameTooltip.SetHyperlink then
			GameTooltip:SetHyperlink(GetSpellLink(info.title))
		else
			local _, ob = GetHighestSpellIndex(info.id)
			if not ob then
				GameTooltip:SetSpellBookItem(info.id, BOOKTYPE_SPELL);
			end
		end
		local skillRank, skillMaxRank = GetProfLevel(tmpSkillName)
		if (skillRank and skillMaxRank and skillName) then
			skillDiffPercent = (skillRank / skillMaxRank) * 100;
			skillColor = '';
			if (skillDiffPercent <= 25 ) then
				skillColor = 'ff0000';
			elseif (skillDiffPercent > 25 and skillDiffPercent <= 50) then
				skillColor = 'ff7f00';
			elseif (skillDiffPercent > 50 and skillDiffPercent <= 75) then
				skillColor = 'ffff00';
			elseif (skillDiffPercent > 75) then
				skillColor = '00ff00';
			end
			if not skillColor then
				skillColor = 'ff0000';
			end
--			DEFAULT_CHAT_FRAME:AddMessage('TradesBar: skillColor = "'..tostring(skillColor)..'", spellID = "'..tostring(self.spellID)..'", skillName = "'..tostring(skillName)..'", skillMaxRank = "'..tostring(skillMaxRank)..'", skillRank = "'..tostring(skillRank)..'"');
			TradesBar_TooltipRanksText[skillName] = {
				left = '|cffffffffSkill Level:|r';
				right = '|cff'..tostring(skillColor)..skillRank..'|r|cffffffff/'..skillMaxRank..'|r';
			}
			GameTooltip:AddLine(" ")
			GameTooltip:AddDoubleLine('|cffffffffSkill Level:|r', '|cff'..skillColor..skillRank..'|r|cffffffff/'..skillMaxRank..'|r')
		end
		GameTooltip:Show();
	elseif info.type == 'item' and info.title then
		if not info.itemLink then
			_, info.itemLink = GetItemInfo(info.title)
		end
		if info.itemLink then
			GameTooltip:SetHyperlink(info.itemLink);
			GameTooltip:Show();
		end
	elseif info.type == 'macro' and info.tooltiptext then
		GameTooltip:SetText(info.tooltiptext)
		if EGDebugger and EGDebugger:AddonRegistered(aName) then
			local macro = SecureButton_GetModifiedAttribute(self, "macro", 'LeftButton')
			local macrotext = SecureButton_GetModifiedAttribute(self, "macrotext", 'LeftButton')
			GameTooltip:AddLine(tostring(macrotext).."\n"..tostring(macro))
		end
		GameTooltip:Show()
	end
end

function TradesBar_ProperSkillName(skillName)
	if (skillName == TRADES_DISENCHANT) then
		skillName = TRADES_ENCHANTING
	elseif (skillName == TRADES_SMELTING or skillName == TRADES_FIND_MINERALS) then
		skillName = TRADES_MINING
	elseif (skillName == TRADES_FIND_HERBS) then
		skillName = TRADES_HERBALISM
	elseif (skillName == TRADES_BASIC_CAMPFIRE) then
		skillName = TRADES_COOKING
	elseif (skillName == TRADES_PICK_LOCK or skillName == TRADES_PICK_POCKET) then
		skillName = TRADES_LOCKPICKING
	elseif (skillName == TRADES_PROSPECTING) then
		skillName = TRADES_JEWELCRAFTING
	elseif (skillName == TRADES_SHADOWEAVE_TAILORING or skillName == TRADES_SPELLFIRE_TAILORING) then
		skillName = TRADES_TAILORING
	elseif (skillName == TRADES_GNOME_ENGINEERING or skillName == TRADES_GOBLIN_ENGINEERING) then
		skillName = TRADES_ENGINEERING
	elseif (skillName == TRADES_DRAGONSCALE_LEATHERWORKING or skillName == TRADES_ELEMENTAL_LEATHERWORKING or skillName == TRADES_TRIBAL_LEATHERWORKING) then
		skillName = TRADES_LEATHERWORKING
	elseif (skillName == TRADES_MILLING) then
		skillName = TRADES_INSCRIPTION
	elseif (skillName == TRADES_SURVEY) then
		skillName = TRADES_ARCHAEOLOGY
	end
	return skillName;
end

-- CONFIG

function TradesBarConfig_OnHide(this)
	if original_heights['TradesBarConfigPanel4'] and original_heights['TradesBarConfigPanel5'] and original_heights['TradesBarConfig'] then
		TradesBarConfigPanel4:SetHeight(original_heights['TradesBarConfigPanel4']);
		TradesBarConfigPanel5:SetHeight(original_heights['TradesBarConfigPanel5']);
		TradesBarConfig:SetHeight(original_heights['TradesBarConfig']);
	end
end

local emptyTable = {}

local function populate(scrollFrame)
	local k = scrollFrame:GetID()
	scrollFrame.v = AdminPanels[k] or emptyTable
	local items
	if scrollFrame.v then
		items = #scrollFrame.v
	else
		items = 0
	end
	FauxScrollFrame_Update(scrollFrame, items, TRADESBAR_ADMIN_MAX_BUTTONS, 16, nil, nil, nil, nil, nil, nil, 1)
	for id = 1, TRADESBAR_ADMIN_MAX_BUTTONS do
		local index = id + FauxScrollFrame_GetOffset(scrollFrame)
		local skillTable = scrollFrame.v[index]
		local btnName = "TradesBarListFramePanel" .. k .. "Button" .. id
		local prevBtnName = "TradesBarListFramePanel" .. k .. "Button" .. (id - 1)
		local button = _G[btnName] 
		if not button then
			button = CreateFrame("Button", btnName, scrollFrame, "TradesBarListFrameButtonTemplate")
		end
		local text = button.text.label
		text:SetText("")
		button:SetID(k)
		button:SetScript("OnClick", nil)
		button:EnableMouse(false)
		if not skillTable then
			break
		end
		if not skillTable.title then
			skillTable.title = UNKNOWN
		end
		if not TradesBarData["skills"][skillTable.title] then
			TradesBarData["skills"][skillTable.title] = {}
		end
		if (TradesBarData["skills"][skillTable.title].visible == 'true') then
			visible = true;
		else
			visible = false;
		end
		button:ClearAllPoints()
		if id == 1 then
			button:SetPoint("TOPLEFT", scrollFrame, "TOPLEFT", 2, -8)
		else
			button:SetPoint("TOPLEFT", _G[prevBtnName], "BOTTOMLEFT")
		end
		-- Default is enabled, but if its disabled and you aquire the skill,
		-- the CheckButton will remain disabled until you relog unless you reenable it like self.
		local r, g, b, a
		if skillTable.subtype or ((skillTable.type and (skillTable.type == 'item' or skillTable.type == 'macro')) or TradesBar_HasSkill(skillTable.title)) then
			r, g, b, a = 0, 1, 0, 1 -- :Enable()
			if not visible then
				r, g, b, a = .8, .8, .8, 1
			end
			button:SetScript("OnClick", TradesBarConfigCheckBox_ToggleSkill)
			button:EnableMouse(true)
		else
			r, g, b, a = .8, .8, .8, .3 -- :Disable()
		end
		local itemName
		if skillTable.type == 'item' then
			itemName = GetItemInfo(skillTable.title)
			button.isItem = 1
		else
			itemName = skillTable.title
			button.isItem = nil
		end
		text:SetTextColor(r, g, b)
		button:SetAlpha(a)
		text:SetText(itemName)
	end
end

-- OnLoad
function TradesBarConfig_OnLoad(self)
	for k = 4, 5 do
		local scrollFrame = TradesBarConfig["panel"..k].scrollFrame
		scrollFrame:SetID(k)
		scrollFrame:SetScript("OnVerticalScroll", function(self, offset)
			FauxScrollFrame_OnVerticalScroll(self, offset, 16, populate)
		end)
		scrollFrame.v = scrollFrame.v or AdminPanels[k]
		populate(scrollFrame)
	end
end

-- OnShow
function TradesBarConfig_OnShow(self)
	TradesBarConfig_OnHide(self)
	-- Top Panel
	local order, hidebar, horizontal;
	if (TradesBarData["settings"]["order"] == 'za') then
		order = true;
	else
		order = false;
	end
	if (TradesBarData["settings"]["hidebar"] == 'true') then
		hidebar = true;
	else
		hidebar = false;
	end
	if (TradesBarData["settings"]["horizontal"] == 'true') then
		horizontal = true;
	else
		horizontal = false;
	end
	TradesBarConfigPanel1CheckButton1:SetChecked( hidebar );
	TradesBarConfigPanel1CheckButton2:SetChecked( horizontal );
	TradesBarConfigPanel1CheckButton3:SetChecked( order );
	TradesBarConfigPanel1SliderScale:SetValue( TradesBarData["settings"]["scale"] or TradesBar:GetEffectiveScale() );
	self.offset = self.offset or 0
	TradesBarConfig_OnLoad(self)
end

function TradesBarConfigScale_OnClick(self)
	TradesBarData["settings"]["scale"] = self:GetValue();
	TradesBar_Scale();
end

function TradesBarConfigCheckBoxHide_OnClick()
	if (TradesBarData["settings"]["hidebar"] == "true") then
		TradesBarData["settings"]["hidebar"] = "false";
	else
		TradesBarData["settings"]["hidebar"] = "true";
	end
	TradesBar_Toggle();
	TradesBar_Update();
end

function TradesBarConfigToggleAll_OnClick(ctype)
	local hideall = "true"
	for _, tab in pairs(Trades) do
		if tab.mountOrPet and tab.mountOrPet == ctype then
			if (TradesBarData["skills"][tab.title].visible == "true") then
				TradesBarData["skills"][tab.title].visible = "false";
			else
				hideall = "false"
				TradesBarData["skills"][tab.title].visible = "true";
			end
		end
	end
	TradesBarData["settings"]["hide_all_"..ctype] = hideall
	TradesBar_Update();
	TradesBar_ScanPets();
	TradesBarConfig_OnShow(TradesBarConfig)
end

function TradesBarConfigDisableAll_OnClick(ctype)
	for _, tab in pairs(Trades) do
		if tab.mountOrPet and tab.mountOrPet == ctype then
			TradesBarData["skills"][tab.title].visible = "false";
		end
	end
	TradesBarData["settings"]["hide_all_"..ctype] = "true"
	TradesBar_Update();
	TradesBarConfig_OnShow(TradesBarConfig)
end

function TradesBarConfigCheckBox_OnClick()
	if (TradesBarData["settings"]["horizontal"] == "true") then
		TradesBarData["settings"]["horizontal"] = "false";
	else
		TradesBarData["settings"]["horizontal"] = "true"
	end
	TradesBar_O();
end

function TradesBarConfigCheckBoxOrder_OnClick()
	if (TradesBarData["settings"]["order"] == "az") then
		TradesBarData["settings"]["order"] = "za";
	else
		TradesBarData["settings"]["order"] = "az"
	end
	TradesBar_O();
end

function TradesBarConfigCheckBox_ToggleSkill(self, button)
	TradesBarData["skills"] = TradesBarData["skills"] or {}
	local text = self.text.label
	local skill = text:GetText();
	if self.isItem then
		local itemName, itemLink = GetItemInfo(skill)
		if itemName then
			local itemString = strmatch(itemLink, "^|c%x+|H(.+)|h%[.+%]")
			local itemId = strmatch(itemString, "item:(%d+):.+")
			skill = "item:" .. itemId
		end
	end
	if self:GetID() == 5 and button == "RightButton" then -- removing custom buttons
		TradesBarData["skills"][skill].visible = "true" -- it's a toggle, make it visible first.
		for k, v in pairs(AdminPanels[5]) do
			if v.title == skill then
				AdminPanels[5][k] = nil
			end
		end
		for k, v in pairs(Trades) do
			if v.title == skill then
				if v.button then
					v.button:Hide()
				end
				Trades[k] = nil
			end
		end
		for k, v in pairs(TradesBarCustomSpells) do
			if v.title == skill then
				TradesBarCustomSpells[k] = nil
			end
		end
		TradesBar_Load()
		TradesBarConfig_OnLoad(TradesBarConfig)
	end
	local checked;
	local hideall = "true"
-- 	DEFAULT_CHAT_FRAME:AddMessage('Toggling '..skill);
	TradesBarData["skills"][skill] = TradesBarData["skills"][skill] or {}
	if (TradesBarData["skills"][skill].visible == "true") then
		TradesBarData["skills"][skill].visible = "false";
		checked = false;
	else
		TradesBarData["skills"][skill].visible = "true";
		hideall = "false"
		checked = true;
	end
	local r, g, b, a
	r, g, b, a = 0, 1, 0, 1 -- :Enable()
	if not checked then
		r, g, b, a = .8, .8, .8, 1
	end
	text:SetTextColor(r, g, b)
	self:SetAlpha(a)
	if TradesBarData["skills"][skill].buttonType then
		TradesBarData["settings"]["hide_all_"..TradesBarData["skills"][skill].buttonType] = hideall
		TradesBar_ScanPets();
	end
	TradesBar_Update();
end

function TradesBarDump()
	TradesBarSkills_Table = AdminPanels
end

function TradesBar_GetSkills()
	for i = 1, 100000 do
		local spellName = GetSpellInfo(i);
		table.insert(TradesBarLocale, {
			['spellID'] = spellID, ['spellName'] = spellName,
		})
	end
end

function TradesBar_GetSpellIDByName(name)
	for i = 1, 100000 do
		if name == GetSpellInfo(i) then
			return i
		end
	end
end

-- Misc script handlers

function TradesBar_OnMouseDown(self, button, ...)
	if IsControlKeyDown() then
		TradesBar_SlashHandler('move')
	else
		self:StartMoving()
	end
end

function TradesBar_OnEnter(self, ...)
	GameTooltip_SetDefaultAnchor(GameTooltip, self)
	GameTooltip:AddLine("Drag to move.")
	GameTooltip:AddLine("Control-click to reenable.")
	GameTooltip:Show()
end

function TradesBarButtonTemplate_OnMouseDown(self, button, ...)
	if (button == "RightButton") then
		if (IsControlKeyDown()) then
			TradesBar_SlashHandler('move')
		elseif (IsShiftKeyDown()) then
			TradesBar_SlashHandler('move', 'Disable')
			TradesBar_SlashHandler('config');
		elseif (IsAltKeyDown()) then
			TradesBar_SlashHandler('move', 'Disable')
			TradesBarBindingFrame.buttonName = self:GetName()
			TradesBarBindingFrameText1:SetText(self.title)
			TradesBarBindingFrame.title = self.title
			TradesBarBindingFrameText2:SetText(TradesBar_ImplodeList(",", GetBindingKey("CLICK " .. self:GetName() .. ":LeftButton")))
			TradesBarBindingFrame:Show()
		end
	end
end

function TradesBarButtonTemplate_OnMouseUp(self, button, ...)
	local f = self:GetParent() 
	f:StopMovingOrSizing();
	self:StopMovingOrSizing();
end

function TradesBarButtonTemplate_OnLeave(self)
	GameTooltip:Hide();
end

function TradesBarBindingFrameCancel_OnClick(self)
	if not SetBinding(self:GetParent().oldKey, self:GetParent().current) then
		_G[self:GetName().."Text4"]:SetText("|cffff0000Binding failed.|r")
		return
	end
	_G[self:GetParent():GetName().."Text2"]:SetText(TradesBar_ImplodeList(",", GetBindingKey("CLICK " .. self:GetParent().buttonName .. ":LeftButton")))
	_G[self:GetParent():GetName().."Text3"]:SetText("Rebound " .. self:GetParent().oldKey .. " to " .. self:GetParent().current)
	_G[self:GetParent():GetName().."Text4"]:SetText("")
	SaveBindings(2)
	self:Disable()
end

function TradesBarBindingFrameClose_OnClick(self)
	-- PlaySound("gsTitleOptionOK");
	self:GetParent():Hide();
end

function TradesBarBindingFrame_OnShow(self)
	if InCombatLockdown() then
		self:Hide()
		return
	end
	self:EnableKeyboard(true);
	_G[self:GetName().."Text3"]:SetText("")
	_G[self:GetName().."Text4"]:SetText("")
	_G[self:GetName().."Cancel"]:Disable()
end

function TradesBarBindingFrame_OnUpdate(self, elapsed)
	if TradesBar_GetSkillLineID and self.title then
		local skillLineID, skillRank, skillMaxRank = TradesBar_GetSkillLineID(self.title), nil, nil;
		if skillLineID then
			_, _, _, skillRank, _, _, skillMaxRank = GetSkillLineInfo(skillLineID);
			if skillRank then
				_G[self:GetName() .. "Count"]:SetText(skillRank)
			end
		end
	end
end

function TradesBarBindingFrame_OnHide(self)
	self:EnableKeyboard(false);
	TradesBar_SlashHandler('move', 'Enable')
end

function TradesBarBindingFrame_OnKeyDown(self, button, ...)
	local mod = strsub(button, 2)
	local modifier = ""
	if IsControlKeyDown() then
		modifier = "CTRL"
	end
	if IsShiftKeyDown() then
		if modifier ~= "" then
			modifier = modifier .. "-SHIFT"
		else
			modifier = "SHIFT"
		end
	end
	if IsAltKeyDown() then
		if modifier ~= "" then
			modifier = modifier .. "-ALT"
		else
			modifier = "ALT"
		end
	end
	local key = button
	if modifier ~= "" then
		key = modifier .. "-" .. key
	end
	if key == "ESCAPE" then
		self:Hide()
		return
	end
-- 				DEFAULT_CHAT_FRAME:AddMessage(key)
	if not (mod == "SHIFT" or mod == "ALT" or mod == "CTRL") then
		local current = GetBindingAction(key)
		if not SetBindingClick(key, self.buttonName) then
			_G[self:GetName().."Text4"]:SetText("|cffff0000Binding failed.|r")
			return
		end
		
		if current ~= "" then
			self.current = current
			self.oldKey = key
			if current ~= "CLICK " .. self.buttonName .. ":LeftButton" then
				_G[self:GetName().."Text4"]:SetText("|cffff0000Unbound "..current.."|r")
				_G[self:GetName().."Cancel"]:Enable()
			end
		end
		
		_G[self:GetName().."Text2"]:SetText(TradesBar_ImplodeList(",", GetBindingKey("CLICK " .. self.buttonName .. ":LeftButton")))
		_G[self:GetName().."Text3"]:SetText("Bound " .. key .. " to " .. self.buttonName)
		
		if ( GetBindingByKey(key) == "SCREENSHOT" ) then
			TakeScreenshot();
			return;
		end
--					self:Hide()
		self:EnableKeyboard(false);
		SaveBindings(2)
	end
end
